/* $Header: /ext/yav/catty/xpx/RCS/colset.c,v 1.16 1994/04/02 13:02:20 yav Exp $
 */

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <stdio.h>

#include "xpx.h"
#include "extern.h"

#define COLSETW 64
#define COLSETH 18
#define COLBOXW 16		/* < COLSETW */

MENU colmenu[MAXCOLSET+1];
static Bool coledopened[MAXCOLSET];
static Window coledwin[MAXCOLSET];
static MENU *coledmn[MAXCOLSET];
static unsigned long rgb_pixel[3];
static GC gc_rgb[3];

char *redraw_col(p)
     MENU *p;
{
  int i, len, w, ww;
  char buf[256];
  GC g;
  COL *cp;

  cp = color_buf+current_pal*MAXCOLSET+p->n;
  if (!(coltbl[p->n] & CM_ALLOCED))
    return "None";
  sprintf(buf, "%x%c%x%x%x", p->n, (coltbl[p->n]&CM_PRIVATE?':':'.'),
	  cp->rgb.green>>4, cp->rgb.red>>4, cp->rgb.blue>>4);
  len = strlen(buf);
  ww = COLSETW-COLBOXW;
  XFillRectangle(dsp, p->win, colgc[p->n], 0, 0, COLBOXW, COLSETH);
  if (p->mode & MenuModeMaskOn) {
    XFillRectangle(dsp, p->win, gc0, COLBOXW, 0, ww, COLSETH);
    g = gc1;
  } else {
    g = gc0;
  }
  w = XTextWidth(fs0, buf, len);
  XDrawString(dsp, p->win, g, COLBOXW+(ww-w)/2, COLSETH-5, buf, len);
  menu_highlight(p);
  return NULL;
}

set_current_color(n)
{
  int i, oldcol;

  if (n != current_col) {
    oldcol = current_col;
    current_col = n;
    if (oldcol >= 0 && oldcol < MAXCOLSET) {
      colmenu[oldcol].mode &= ~MenuModeMaskOn;
      redraw_window(colmenu[oldcol].win);
    }
    if (n >= 0 && n < MAXCOLSET) {
      colmenu[n].mode |= MenuModeMaskOn;
      redraw_window(colmenu[n].win);
    }
  }
}

int colev(p, ev)
     MENU *p;
     XEvent *ev;
{
  switch(ev->type) {
  case ButtonPress:
    if (double_clicked) {
      if (coledopened[p->n])
	unmap_coled_window(p->n);
      else
	create_coledit_window(p->n);
    } else {
      set_current_color(p->n);
    }
    return 1;
  }
  return 0;
}

get_rgb_gc()
{
  int i;
  XColor col;
  char *name[3] = {"red", "green", "blue"};

  for (i = 0; i < 3; i++) {
    gc_rgb[i] = XCreateGC(dsp, win, 0, 0);
    XSetBackground(dsp, gc_rgb[i], bg);
    XSetForeground(dsp, gc_rgb[i], fg);
    if (parse_color(&col, name[i]) == 0) {
      rgb_pixel[i] = col.pixel;
      XSetForeground(dsp, gc_rgb[i], col.pixel);
    }
  }
}

create_colset_window(x, y)
{
  int i;
  MENU *p;
  
  p = colmenu;
  for (i = 0; i < MAXCOLSET; i++) {
    memset(p, 0, sizeof(*p));
    p->n = i;
    p->y = (COLSETH+BDW)*i;
    p->w = COLSETW;
    p->h = COLSETH;
    p->im = ExposureMask|EnterWindowMask|LeaveWindowMask|ButtonPressMask;
    p->mode = MenuModeMaskBorder;
    p->event = colev;
    p->strfunc = redraw_col;
    p++;
  }
  p->n = -1;
  menuwin_create(colmenu, win, x, y);
  for (i = 0; i < MAXCOLSET; i++)
    colgc[i] = XCreateGC(dsp, colmenu[i].win, 0, 0);
  memset(coledmn, 0, sizeof(coledmn));
  memset(coledopened, False, sizeof(coledopened));
  get_rgb_gc();
  return 0;
}

#define COLEDX	400
#define COLEDY	50
#define COLEDDX	20
#define COLEDDY	32

#define COLBARH	16
#define COLEDW	380
#define COLEDH	(COLBARH*3+BDW*2)
#define COLED_COLOR	0
#define COLED_NRED	1
#define COLED_NGREEN	2
#define COLED_NBLUE	3
#define COLED_RED	4
#define COLED_GREEN	5
#define COLED_BLUE	6
#define COLED_OK	7
#define IMASK	\
(ExposureMask|ButtonPressMask|EnterWindowMask|LeaveWindowMask|ButtonMotionMask)

int coledev();
char *coledstr();
static MENU coledmenu[] = {
  {
    COLED_COLOR,	MenuModeMaskBorder,
    2, 2, 32, COLEDH-BDW*2-4,
    0, ExposureMask|ButtonPressMask|EnterWindowMask|LeaveWindowMask, 0,
    coledev, NULL, 0, 0, ""},
  {
    COLED_NRED,	MenuModeMaskBorder,
    48, 0, 32, COLBARH-BDW,
    0, ExposureMask, 0,
    NULL, coledstr, 0, 8, NULL},
  {
    COLED_NGREEN,	MenuModeMaskBorder,
    48, COLBARH, 32, COLBARH-BDW,
    0, ExposureMask, 0,
    NULL, coledstr, 0, 8, NULL},
  {
    COLED_NBLUE,	MenuModeMaskBorder,
    48, COLBARH*2, 32, COLBARH-BDW,
    0, ExposureMask, 0,
    NULL, coledstr, 0, 8, NULL},
  {
    COLED_RED,	MenuModeMaskBorder,
    48+32, 0, 255+8, COLBARH-BDW,
    0, IMASK, 0,
    coledev, NULL, 0, 0, ""},
  {
    COLED_GREEN,	MenuModeMaskBorder,
    48+32, COLBARH, 255+8, COLBARH-BDW,
    0, IMASK, 0,
    coledev, NULL, 0, 0, ""},
  {
    COLED_BLUE,	MenuModeMaskBorder,
    48+32, COLBARH*2, 255+8, COLBARH-BDW,
    0, IMASK, 0,
    coledev, NULL, 0, 0, ""},
  {
    COLED_OK,	MenuModeMaskBorder,
    COLEDW-(32+BDW*2)-2, COLEDH-(20+BDW*2)-2, 32, 20,
    0, ExposureMask|ButtonPressMask|EnterWindowMask|LeaveWindowMask,
    SouthEastGravity,
    coledev, NULL, 0, 0, "OK"},
  {
    -1,	0,
    0, 0, 0, 0,
    0, 0, 0,
    NULL, NULL, 0, 0, NULL}
};

static Window colorbarwin[MAXCOLSET][3];

create_coledit_window(n)
{
  XSizeHints hint;
  MENU *p;
  int i;
  char buf[256];

  if (n >= MAXCOLSET) {
    error("coledit: color %d!", n);
    return;
  }
  if (coledwin[n]) {
    XMapWindow(dsp, coledwin[n]);
    coledopened[n] = True;
    return;
  }
  hint.x = COLEDX+COLEDDX*n;
  hint.y = COLEDY+COLEDDY*n;
  hint.width = COLEDW;
  hint.height = COLEDH;
  hint.flags = USPosition|USSize;
  coledwin[n] = XCreateSimpleWindow(dsp, DefaultRootWindow(dsp),
				    hint.x, hint.y, hint.width, hint.height,
				    border_width, border_color, bg);
  sprintf(buf, "%s color %d", myname, n);
  XSetStandardProperties(dsp, coledwin[n], buf, buf, None, NULL, NULL, &hint);
  XSelectInput(dsp, coledwin[n], ExposureMask);
  XMapRaised(dsp, coledwin[n]);
  p = (MENU *)malloc(sizeof(coledmenu));
  if (p == NULL) {
    error("more core!");
    return;
  }
  memcpy(p, coledmenu, sizeof(coledmenu));
  p->cursor = n;
  menuwin_create(p, coledwin[n], 0, 0);
  coledmn[n] = p;
  for (i = 0; i < 3; i++) {
    static int ntbl[3] = {COLED_RED, COLED_GREEN, COLED_BLUE};
    colorbarwin[n][i] = XCreateSimpleWindow(dsp, (p+ntbl[i])->win,
					    BDW, BDW, 100, COLBARH-BDW*3,
					    0, fg, rgb_pixel[i]);
    XSelectInput(dsp, colorbarwin[n][i], ExposureMask|EnterWindowMask|LeaveWindowMask);
    XMapWindow(dsp, colorbarwin[n][i]);
  }
  coledopened[n] = True;
}

unmap_coled_window(n)
{
  XUnmapWindow(dsp, coledwin[n]);
  coledopened[n] = False;
}

update_coled_sub(n)
{
  MENU *p;

  if (coledwin[n]) {
    redraw_window(coledwin[n]);
    p = coledmn[n];
    while (p->n >= 0) {
      redraw_window(p->win);
      p++;
    }
  }
}

update_coled_window()
{
  int i;

  for (i = 0; i < MAXCOLSET; i++)
    update_coled_sub(i);
}

update_colmenu(n)
{
  update_coled_sub(n);
  redraw_window(colmenu[n].win);
}

/* sub.c */
extern MENU *get_top_menu();

#define coled_num(p)	(get_top_menu(p)->cursor)
#define coled_colnum(p)	(current_pal*MAXCOLSET+coled_num(p))
#define coled_col(p)	(color_buf+coled_colnum(p))

char *coledstr(p)
     MENU *p;
{
  int i;
  
  switch(p->n) {
  case COLED_NRED:
    i = coled_col(p)->rgb.red;
    break;
  case COLED_NGREEN:
    i = coled_col(p)->rgb.green;
    break;
  case COLED_NBLUE:
    i = coled_col(p)->rgb.blue;
    break;
  }
  sprintf(p->str, "%d", i);
  return p->str;
}

draw_rgb_bar(p)
     MENU *p;
{
  int i, w;
  COL *col;

  col = coled_col(p);
  switch(p->n) {
  case COLED_RED:
    i = 0;
    w = col->rgb.red;
    break;
  case COLED_GREEN:
    i = 1;
    w = col->rgb.green;
    break;
  case COLED_BLUE:
    i = 2;
    w = col->rgb.blue;
    break;
  }
#if 1
  XClearWindow(dsp, p->win);
  XResizeWindow(dsp, colorbarwin[coled_num(p)][i], w<1?1:w, COLBARH-BDW*3);
  XFillRectangle(dsp, p->win, gc0, BDW+w, BDW, 1, p->h-BDW*2);
#else
  XFillRectangle(dsp, p->win, gc_rgb[i], 0, 0, w, p->h);
  XFillRectangle(dsp, p->win, gc0, w, 0, 1, p->h);
#endif
}

int coledev(p, ev)
     MENU *p;
     XEvent *ev;
{
  int i;
  XColor col;

  switch(ev->type) {
  case Expose:
    switch(p->n) {
    case COLED_COLOR:
      XFillRectangle(dsp, p->win, colgc[p->cursor],
		     ev->xexpose.x, ev->xexpose.y,
		     ev->xexpose.width, ev->xexpose.height);
      break;
    case COLED_RED:
    case COLED_GREEN:
    case COLED_BLUE:
      if (ev->xexpose.count == 0)
	draw_rgb_bar(p);
      break;
    }
    break;
  case ButtonPress:
    switch(p->n) {
    case COLED_OK:
      unmap_coled_window(coled_num(p));
      return 1;
    case COLED_COLOR:
      XRaiseWindow(dsp, coledwin[coled_num(p)]);
      return 1;
    }
  case MotionNotify:
    if (!(coltbl[coled_num(p)] & CM_PRIVATE))
      return 1;
    switch(ev->type) {
    case MotionNotify:
      i = ev->xmotion.x;
      break;
    case ButtonPress:
      i = ev->xbutton.x;
      break;
    default:
      fprintf(stderr, "unknown colev %d\n", ev->type);
      break;
    }
    if (i < 0)
      i = 0;
    if (i > 255)
      i = 255;
    switch(p->n) {
    case COLED_RED:
      coled_col(p)->rgb.red = i;
      redraw_window(get_top_menu(p)[COLED_NRED].win);
      break;
    case COLED_GREEN:
      coled_col(p)->rgb.green = i;
      redraw_window(get_top_menu(p)[COLED_NGREEN].win);
      break;
    case COLED_BLUE:
      coled_col(p)->rgb.blue = i;
      redraw_window(get_top_menu(p)[COLED_NBLUE].win);
      break;
    }
    calcol(&col, coled_col(p));
    col.pixel = pixtbl[coled_num(p)];
    XStoreColor(dsp, DefaultColormap(dsp, scr), &col);
    redraw_window(colmenu[coled_num(p)].win);
    redraw_window(p->win);
    return 1;
  }
  return 0;
}

/* End of file */
