/* $Header: /ext/yav/catty/xpx/RCS/file.c,v 1.26 1994/04/03 15:53:35 yav Exp $
 */

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>
#include <stdio.h>

#include "xpx.h"
#include "extern.h"
#include "icondef.h"
#include "infodef.h"

#define FILEWINWIDTH	320
#define FILEWINHEIGHT	320

#define MAXFILE	1024
#define MAXFILENAME	64

#define PATHX	8
#define PATHY	8
#define PATHW	300
#define PATHH	24

#define NAMEX	160
#define NAMEY	100
#define NAMEW	140
#define NAMEH	PATHH

#define DIRSLOTMAX	12

#define SCROLLX		(8-BDW)
#define SCROLLY		40
#define SCROLLW		128
#define SCROLLH		20
#define DIRSLOTX	(SCROLLX+BDW)
#define DIRSLOTY	(SCROLLY+SCROLLH+BDW*2)
#define DIRSLOTW	SCROLLW
#define DIRSLOTH	18

#define FILEMENUX	196
#define FILEMENUY	160
#define FILEMENUW	64
#define FILEMENUH	20

#define COLNMX	PALSETX
#define COLNMY	(PALSETY-24)
#define COLNMW	112
#define COLNMH	16
#define IMGNMX	COLNMX+COLNMW+8
#define IMGNMY	COLNMY
#define IMGNMW	COLNMW
#define IMGNMH	COLNMH

static Window filewin;
static int select_slot;
static char filenamebuf[MAXFILE*MAXFILENAME];
static char current_path[MAXPATHLEN];
static char current_name[MAXFILENAME];
static char noname[] = "*Noname*";
static int nametop;

extern MENU infomenu[];		/* info.c */

extern int rd_ccb();
extern int rd_cbb();
extern int rd_csb();
extern int rd_cxb();
extern int rd_pnm();
extern int rd_kcf();
extern int wr_kcf();
extern int rd_cel();
extern int wr_cel();
extern int rd_xbm();
extern int wr_xbm();
extern int wr_ppm();

#define F_IMG	(1<<0)
#define F_COL	(1<<1)
#define F_ALL	(F_IMG|F_COL)

typedef struct {
  char suffix[16];
  short mode;
  int (*readroutine)();
  int (*writeroutine)();
  char decoder[256];
  char encoder[256];
} FTYPE;

#define FTYPEMAX 32
static FTYPE *file_type_tbl;
static FTYPE default_file_type[] = {
  {".ccb",	F_COL,	rd_ccb,	NULL,	"",		""},
  {".cbb",	F_IMG,	rd_cbb,	NULL,	"",		""},
  {".csb",	F_IMG,	rd_csb,	NULL,	"",		""},
  {".pbm",	F_ALL,	rd_pnm,	NULL,	"",		""},
  {".pgm",	F_ALL,	rd_pnm,	NULL,	"",		""},
  {".ppm",	F_ALL,	rd_pnm,	wr_ppm,	"",		""},
  {".mag",	F_ALL,	rd_pnm,	NULL,	"magtoppm",	""},
  {".kcf",	F_COL,	rd_kcf,	wr_kcf, "",		""},
  {".cel",	F_IMG,	rd_cel,	wr_cel, "",		""},
  {".xbm",	F_IMG,	rd_xbm,	wr_xbm,	"",		""},	
  {".gif",	F_ALL,	rd_pnm,	wr_ppm,	"giftoppm",	"ppmtogif"},	
  {"",		0,	NULL,	NULL,	"",		""}
};

#define FILE_PATH	0
#define FILE_NAME	1
#define FILE_OPEN	2
#define FILE_SAVE	3
#define FILE_DUMMY1	4
#define FILE_REMOVE	5
#define FILE_DUMMY2	6
#define FILE_OK		7

int fileev();
int slotev();
int slotev2();

#define IMASK (ExposureMask|EnterWindowMask|LeaveWindowMask|ButtonPressMask)
static MENU filemenu[] = {
  {
    FILE_PATH,	MenuModeMaskBorder,
    PATHX, PATHY, PATHW, PATHH,
    0, IMASK|KeyPressMask, 0,
    fileev, NULL, 0, MAXPATHLEN, NULL},
  {
    FILE_NAME,	MenuModeMaskBorder,
    NAMEX, NAMEY, NAMEW, NAMEH,
    0, IMASK|KeyPressMask, 0,
    fileev, NULL, 0, 64, NULL},
  {
    FILE_OPEN,	MenuModeMaskBorder,
    FILEMENUX, FILEMENUY+(FILEMENUH+BDW)*0, FILEMENUW, FILEMENUH,
    0, IMASK, 0,
    fileev, NULL, 0, 0, "Open"},
  {
    FILE_SAVE,	MenuModeMaskBorder,
    FILEMENUX, FILEMENUY+(FILEMENUH+BDW)*1, FILEMENUW, FILEMENUH,
    0, IMASK, 0,
    fileev, NULL, 0, 0, "Save"},
  {
    FILE_DUMMY1,	MenuModeMaskBorder,
    FILEMENUX, FILEMENUY+(FILEMENUH+BDW)*2, FILEMENUW, FILEMENUH,
    0, ExposureMask, 0,
    NULL, NULL, 0, 0, "----"},
  {
    FILE_REMOVE,	MenuModeMaskBorder,
    FILEMENUX, FILEMENUY+(FILEMENUH+BDW)*3, FILEMENUW, FILEMENUH,
    0, IMASK, 0,
    fileev, NULL, 0, 0, "Remove"},
  {
    FILE_DUMMY2,	MenuModeMaskBorder,
    FILEMENUX, FILEMENUY+(FILEMENUH+BDW)*4, FILEMENUW, FILEMENUH,
    0, ExposureMask, 0,
    NULL, NULL, 0, 0, "----"},
  {
    FILE_OK,	MenuModeMaskBorder,
    FILEMENUX, FILEMENUY+(FILEMENUH+BDW)*5, FILEMENUW, FILEMENUH,
    0, IMASK, 0,
    fileev, NULL, 0, 0, "OK"},
  {
    -1,	0,
    0, 0, 0, 0,
    0, 0, 0,
    NULL, NULL, 0, 0, NULL}};

static MENU slotwin[DIRSLOTMAX+2+1];

#define NAMEMENU_IMAGE	0
#define NAMEMENU_COLOR	1

char *namemenu_str();

static MENU namemenu[] = {
  {
    NAMEMENU_IMAGE,	MenuModeMaskBorder,
    IMGNMX, IMGNMY, IMGNMW, IMGNMH,
    0, ExposureMask, 0,
    NULL, namemenu_str, 0, 0, NULL},
  {
    NAMEMENU_COLOR,	MenuModeMaskBorder,
    COLNMX, COLNMY, COLNMW, COLNMH,
    0, ExposureMask, 0,
    NULL, namemenu_str, 0, 0, NULL},
  {
    -1,	0,
    0, 0, 0, 0,
    0, 0, 0,
    NULL, NULL, 0, 0, NULL}};

char *namemenu_str(p)
     MENU *p;
{
  switch(p->n) {
  case NAMEMENU_IMAGE:
    return filename(imgfilename);
  case NAMEMENU_COLOR:
    return filename(colfilename);
  }
  /* NOT REACHED */
}

/*
 * return number of file (0 is error!)
 */
get_filenames()
{
  int i, n;
  FILE *fp;
  int fd[2];
  char buf[MAXPATHLEN+1];
  
  i = select_slot - nametop;
  if (0 <= i && i < DIRSLOTMAX)
    slotwin[i].mode &= ~MenuModeMaskOn;
  select_slot = nametop = 0;
  slotwin[0].mode |= MenuModeMaskOn;
  n = get_dir(current_path, filenamebuf, MAXFILENAME, MAXFILE);
  for (i = n; i < MAXFILE; i++)
    filenamebuf[i*MAXFILENAME] = '\0';
  redraw_window(filemenu[FILE_PATH].win);
  for (i = 0; i < DIRSLOTMAX; i++)
    redraw_window(slotwin[i].win);
  return n;
}

char *slotwin_strfunc(p)
     MENU *p;
{
  return filenamebuf+(nametop+p->n)*MAXFILENAME;
}

create_file_window()
{
  int i, x, y;
  XSizeHints hint;
  MENU *p;
  char buf[256];
  
  get_drawable_size(DefaultRootWindow(dsp), &x, &y);
  hint.x = (x-FILEWINWIDTH)/2;
  hint.y = (y-FILEWINHEIGHT)/2;
  hint.width = FILEWINWIDTH;
  hint.height = FILEWINHEIGHT;
  hint.flags = USPosition|USSize;
  filewin = XCreateSimpleWindow(dsp, DefaultRootWindow(dsp), hint.x, hint.y,
				hint.width, hint.height,
				border_width, border_color, bg);
  sprintf(buf, "%s file", myname);
  XSetStandardProperties(dsp, filewin, buf, buf, None, NULL, NULL, &hint);
  menuwin_create(filemenu, filewin, 0, 0);
  
  x = DIRSLOTX;
  y = DIRSLOTY;
  p = slotwin;
  for (i = 0; i < DIRSLOTMAX; i++) {
    memset(p, 0, sizeof(*p));
    p->n = i;
    p->x = x;
    p->y = y;
    p->w = DIRSLOTW;
    p->h = DIRSLOTH;
    p->mode = MenuModeMaskLeft;
    p->im = ExposureMask|EnterWindowMask|LeaveWindowMask|ButtonPressMask;
    p->event = slotev;
    p->strfunc = slotwin_strfunc;
    p->strlen = 64;
    p++;
    y += DIRSLOTH;
  }
  x = SCROLLX;
  y = SCROLLY;
  while (i < DIRSLOTMAX+2) {
    memset(p, 0, sizeof(*p));
    p->n = i;
    p->x = x;
    p->y = y;
    p->w = SCROLLW;
    p->h = SCROLLH;
    p->mode = MenuModeMaskBorder;
    p->im = ExposureMask|EnterWindowMask|LeaveWindowMask|
      ButtonPressMask|ButtonReleaseMask;
    p->event = slotev2;
    y += SCROLLH+BDW*2+DIRSLOTH*DIRSLOTMAX;
    p++;
    i++;
  }
  p->n = -1;
  menuwin_create(slotwin, filewin, 0, 0);
  
  change_directory("");		/* get current */
  return 0;
}

map_file_window()
{
  XMapRaised(dsp, filewin);
}

unmap_file_window()
{
  XUnmapWindow(dsp, filewin);
}

close_file_window()
{
  int i;
  
  XDestroyWindow(dsp, filewin);
}

char *get_parent_pathname(path)
     char *path;
{
  char *p;
  extern char *strrchr();
  
  p = strrchr(path, '/');
  if (p != NULL) {
    if (p == path)
      p++;
    *p = '\0';
  }
  return path;
}

change_directory(dir)
     char *dir;
{
  int i;
  char buf[MAXPATHLEN];
  char buf2[MAXPATHLEN];
  
  switch(*dir) {
  case '\0':
    /* current working directory pathname */
    get_current_dir(buf);
    break;
  case '/':
    /* absolute pathname */
    strcpy(buf, dir);
    break;
  default:
    /* relative pathname */
    strcpy(buf, current_path);
    strcpy(buf2, dir);
    i = strlen(buf2);
    if (i && buf2[i-1] == '/')
      buf2[i-1] = '\0';
    if (strcmp(buf2, "..") == 0) {
      get_parent_pathname(buf);
    } else if (strcmp(buf2, ".") != 0) {
      i = strlen(buf);
      if (i && buf[i-1] != '/')
	strcat(buf, "/");
      strcat(buf, buf2);
    }
  }
  if (get_dir(buf, buf2, sizeof(buf2), 1)) {
    strcpy(current_path, buf);
    strcpy(filemenu[FILE_PATH].str, buf);
    filemenu[FILE_PATH].cursor = strlen(filemenu[FILE_PATH].str);
    redraw_window(filemenu[FILE_PATH].win);
    get_filenames();
    return 0;
  }
  return 1;
}

void set_imgfilename(name)
     char *name;
{
  strcpy(imgfilename, name);
  redraw_window(namemenu[NAMEMENU_IMAGE].win);
  redraw_window(infomenu[INFOMENU_IMGFILE].win);
}

void set_colfilename(name)
     char *name;
{
  strcpy(colfilename, name);
  redraw_window(namemenu[NAMEMENU_COLOR].win);
  redraw_window(infomenu[INFOMENU_COLFILE].win);
}

create_filename_window()
{
  menuwin_create(namemenu, win, 0, 0);
}

char *skip_space(p)
     char *p;
{
  while (*p && (*p == ' ' || *p == '\t' || *p == '\n'))
    p++;
  return p;
}

/* if not found matching string, return -1 */
int match_string(tbl, p)
     char *tbl[];
     char *p;
{
  int i;
  
  for (i = 0; tbl[i] != NULL; i++) {
    if (strncmp(p, tbl[i], strlen(tbl[i])) == 0)
      return i;
  }
  return -1;			/* not found */
}

char *get_string(str, p, n)
     char str[];
     char *p;
{
  int i;
  
  i = 0;
  if (strncmp(p, "None", 4) == 0) {
    p += 4;
  } else {
    if (*p == '"')
      p++;
    while (*p && *p != '"')
      str[i++] = *p++;
    if (*p == '"')
      p++;
  }
  str[i] = '\0';
  return p;
}

parse_str(tbl, p)
     char *tbl[];
     char **p;
{
  int i;
  char *p1;
  char buf[256];
  
  i = 0;
  if (strncmp(*p, "None", 4) == 0) {
    *p += 4;
  } else {
    i = match_string(tbl, *p);
    if (i < 0) {
      p1 = buf;
      while (**p && **p != ' ' && **p != '\t' && **p != '\n') {
	*p1++ = **p;
	++*p;
      }
      *p1 = '\0';
      error("parse ``%s'' unknown!", buf);
    } else {
      *p += strlen(tbl[i]);
    }
  }
  *p = skip_space(*p);
  return i;
}

parse_ftype(pp, p)
     FTYPE *pp;
     char *p;
{
  int i;
  static char *modestr[] = {"color", "image", "all", NULL};
  static short modetbl[] = {F_COL,   F_IMG,   F_ALL};
  static char *funcstr[] = {"ccb", "cbb", "csb", "cxb",
			      "kcf", "cel",
			      "pnm", "ppm",
			      "xbm",
			      NULL};
  static int (*rdfn[])() = {rd_ccb, rd_cbb, rd_csb, rd_cxb,
			      rd_kcf, rd_cel,
			      rd_pnm, rd_pnm,
			      rd_xbm
			      };
  static char *wfuncstr[] = {"kcf", "cel",
			       "ppm",
			       "xbm",
			       NULL
			       };
  static int (*wrfnc[])() = {wr_kcf, wr_cel,
			       wr_ppm,
			       wr_xbm
			       };
  
  i = 0;
  p = skip_space(p);
  p = get_string(pp->suffix, p, sizeof(pp->suffix));
  if (!*p)
    return 0;
  p = skip_space(p);
  i = parse_str(modestr, &p);
  if (i >= 0)
    pp->mode = modetbl[i];
  pp->readroutine = NULL;
  i = parse_str(funcstr, &p);
  if (i >= 0)
    pp->readroutine = rdfn[i];
  pp->writeroutine = NULL;
  i = match_string(wfuncstr, p);
  i = parse_str(wfuncstr, &p);
  if (i >= 0)
    pp->writeroutine = wrfnc[i];
  pp->decoder[0] = '\0';
  p = get_string(pp->decoder, p, sizeof(pp->decoder));
  p = skip_space(p);
  pp->encoder[0] = '\0';
  p = get_string(pp->encoder, p, sizeof(pp->encoder));
  
  return 1;
}

int read_config_file(pp, name)
     FTYPE *pp;
     char *name;
{
  int i, n;
  Bool f;
  FILE *fp;
  char buf[1024];
  
  n = 0;
  fp = fopen(name, "r");
  if (fp != NULL) {
    f = False;
    while (fgets(buf, sizeof(buf), fp) != NULL) {
      if (*buf == '#')
	continue;
      if (*buf == '*')
	f = False;
      if (strncmp(buf, "*file", 5) == 0) {
	f = True;
      } else {
	if (f) {
	  i = parse_ftype(pp+n, buf);
	  if (!i)
	    error("file type parse error ``%s''", buf);
	  n += i;
	}
      }
    }
    fclose(fp);
  }
  return n;
}

/* return support file type count */
init_file_type()
{
  int i, n;
  Bool f;
  FILE *fp;
  FTYPE *pp;
  char *p;
  char buf[MAXPATHLEN];
  
  n = 0;
  file_type_tbl = (FTYPE *)malloc(sizeof(FTYPE)*FTYPEMAX);
  if (file_type_tbl == NULL) {
    error("more core!");
  } else {
    p = (char *)getenv("HOME");
    if (p != NULL) {
      strcpy(buf, p);
      strcat(buf, "/.xpxrc");
      n += read_config_file(file_type_tbl+n, buf);
    }
    n += read_config_file(file_type_tbl+n, ".xpxrc");
    pp = default_file_type;
    while (n < FTYPEMAX-1)
      file_type_tbl[n++] = *pp++;
    file_type_tbl[n].suffix[0] = '\0';
  }
#if 0
  for (pp = file_type_tbl; pp->suffix[0]; pp++) {
    fprintf(stderr, "ftype [%s] [%ld] [%ld] [%s] [%s]\n",
	    pp->suffix,
	    pp->readroutine, pp->writeroutine,
	    pp->decoder, pp->encoder);
  }
#endif
  return n;
}

read_any_file(p, name)
     FTYPE *p;
     char *name;
{
  FILE *fp;
  int i, r, fd[2];
  
  message("``%s'' reading...", name);
  if (p->decoder[0] != '\0') {
    pipe(fd);
    if (fork() == 0) {
      close(1);
      dup(fd[1]);
      close(fd[1]);
      close(fd[0]);
      execlp(p->decoder, p->decoder, name, NULL);
      exit(0);
    }
    fp = fdopen(fd[0], "r");
    close(fd[1]);
  } else {
    fp = fopen(name, "r");
    if (fp == NULL) {
      message("!open error!\n");
      return 1;
    }
  }
  if (p->mode & F_COL)
    set_colfilename(name);
  if (p->mode & F_IMG)
    set_imgfilename(name);
  r = 1;
  if (p->readroutine != NULL)
    r = p->readroutine(fp);
  fclose(fp);
  if (p->decoder[0] != '\0')
    wait(&i);
  if (r) {
    if (p->mode & F_COL)
      set_colfilename("");
    if (p->mode & F_IMG)
      set_imgfilename("");
    message("!read error!\n");
    return r;
  }
  message("done.\n");
  return 0;
}


write_any_file(p, name)
     FTYPE *p;
     char *name;
{
  FILE *fp;
  int r, i, fd[2];
  
  message("``%s'' writing...", name);
  
  fp = fopen(name, "w");
  if (fp == NULL) {
    message("!open error!\n");
    return 1;
  }
  if (p->encoder[0] != '\0') {
    pipe(fd);
    if (fork() == 0) {
      close(0);
      dup(fd[0]);
      close(1);
      dup(fileno(fp));
      close(fd[1]);
      close(fd[0]);
      execlp(p->encoder, p->encoder, NULL);
      exit(0);
    }
    fp = fdopen(fd[1], "w");
    close(fd[0]);
  }
  r = 1;
  if (p->writeroutine == NULL) {
    message("!\nNot supported format!\nAbort.\n");
    fclose(fp);
    return 1;
  }
  r = p->writeroutine(fp, name);
  fclose(fp);
  if (p->encoder[0] != '\0')
    wait(&i);
  if (r) {
    message("!write error!\n");
    return r;
  }
  message("done.\n");
  return 0;
}

set_current_name(name)
     char *name;
{
  strcpy(current_name, name);
  strcpy(filemenu[FILE_NAME].str, name);
  filemenu[FILE_NAME].cursor = strlen(current_name);
  redraw_window(filemenu[FILE_NAME].win);
}

get_filetype(name)
     char *name;
{
  int i, n;
  
  for (n = 0; file_type_tbl[n].suffix[0] != '\0'; n++) {
    i = strlen(name) - strlen(file_type_tbl[n].suffix);
    if (i >= 0 && strcmp(name+i, file_type_tbl[n].suffix) == 0) {
      return n;
    }
  }
  return -1;
}

read_data_file(name)
     char *name;
{
  int i;
  
  i = get_filetype(name);
  if (i < 0)
    return 1;
  return read_any_file(file_type_tbl+i, name);
}

file_open()
{
  int i;
  char buf[MAXPATHLEN];
  
  i = strlen(current_name);
  if (i && current_name[i-1] == '/') {
    strcpy(buf, current_name);
    buf[i-1] = '\0';
    change_directory(buf);
    set_current_name("");
  }
  if (current_name[0] == '\0')
    return;
  sprintf(buf, "%s/%s", current_path, current_name);
  return read_data_file(buf);
}

file_save()
{
  int i;
  char buf[MAXPATHLEN];
  
  i = strlen(current_name);
  if (!i || current_name[i-1] == '/')
    return 1;
  sprintf(buf, "%s/%s", current_path, current_name);
  i = get_filetype(buf);
  if (i < 0)
    return 1;
  return write_any_file(file_type_tbl+i, buf);
}

file_remove()
{
  int i;
  char buf[MAXPATHLEN];
  
  i = strlen(current_name);
  if (!i || current_name[i-1] == '/')
    return;
  sprintf(buf, "%s/%s", current_path, current_name);
  if (unlink(buf) == 0) {
    for (i = select_slot; i < MAXFILE-1; i++) {
      if (filenamebuf[i*MAXFILENAME] == '\0')
	break;
      strcpy(filenamebuf+i*MAXFILENAME, filenamebuf+(i+1)*MAXFILENAME);
    }
    set_current_name(filenamebuf+(select_slot)*MAXFILENAME);
    for (i = select_slot - nametop; 0 <= i && i < DIRSLOTMAX; i++)
      redraw_window(slotwin[i].win);
  }
}

scroll_down()
{
  int i;
  
  if (nametop == 0)
    return;
  --nametop;
  for (i = DIRSLOTMAX-1; i > 0; --i) {
    if (slotwin[i-1].mode & MenuModeMaskOn)
      slotwin[i].mode |= MenuModeMaskOn;
    else
      slotwin[i].mode &= ~MenuModeMaskOn;
    XCopyArea(dsp, slotwin[i-1].win, slotwin[i].win, gc0, 0, 0,
	      slotwin[i].w, slotwin[i].h, 0, 0);
  }
  if (nametop == select_slot)
    slotwin[0].mode |= MenuModeMaskOn;
  else
    slotwin[0].mode &= ~MenuModeMaskOn;
  redraw_window(slotwin[0].win);
}

scroll_up()
{
  int i;
  
  if (filenamebuf[(nametop+DIRSLOTMAX-1)*MAXFILENAME] == '\0')
    return;
  ++nametop;
  for (i = 0; i < DIRSLOTMAX-1; i++) {
    if (slotwin[i+1].mode & MenuModeMaskOn)
      slotwin[i].mode |= MenuModeMaskOn;
    else
      slotwin[i].mode &= ~MenuModeMaskOn;
    XCopyArea(dsp, slotwin[i+1].win, slotwin[i].win, gc0, 0, 0,
	      slotwin[i].w, slotwin[i].h, 0, 0);
  }
  if (nametop+DIRSLOTMAX-1 == select_slot)
    slotwin[DIRSLOTMAX-1].mode |= MenuModeMaskOn;
  else
    slotwin[DIRSLOTMAX-1].mode &= ~MenuModeMaskOn;
  redraw_window(slotwin[DIRSLOTMAX-1].win);
}

int slotev(p, ev)
     MENU *p;
     XEvent *ev;
{
  int i, oldsel;
  
  switch(ev->type) {
  case ButtonPress:
    oldsel = select_slot;
    select_slot = nametop+p->n;
    set_current_name(filenamebuf+select_slot*MAXFILENAME);
    i = oldsel - nametop;
    if (0 <= i && i < DIRSLOTMAX) {
      slotwin[i].mode &= ~MenuModeMaskOn;
      redraw_window(slotwin[i].win);
    }
    i = select_slot - nametop;
    slotwin[i].mode |= MenuModeMaskOn;
    redraw_window(slotwin[i].win);
    if (double_clicked)
      file_open();
    return 1;
  }
  return 0;
}

int slotev2(p, ev)
     MENU *p;
     XEvent *ev;
{
  switch(ev->type) {
  case Expose:
    switch(p->n) {
    case DIRSLOTMAX:
      menu_icon_center(p, ICON_up);
      break;
    case DIRSLOTMAX+1:
      menu_icon_center(p, ICON_down);
      break;
    }
    return 1;
  }
  return 0;
}

int fileev(p, ev)
     MENU *p;
     XEvent *ev;
{
  switch(ev->type) {
  case ButtonPress:
    switch(p->n) {
    case FILE_OPEN:
      file_open();
      break;
    case FILE_SAVE:
      file_save();
      break;
    case FILE_DUMMY1:
      break;
    case FILE_REMOVE:
      file_remove();
      break;
    case FILE_DUMMY2:
      break;
    case FILE_OK:
      menu_stat_change(1, 0);
      break;
    }
    p->mode &= ~MenuModeMaskOn;	/* every off */
    redraw_window(p->win);
    return 1;
  case KeyPress:
    switch(p->n) {
    case FILE_PATH:
      {
	int i;
	KeySym key;
	char buf[16];
	i = XLookupString((XKeyEvent *)ev, buf, sizeof(buf)-1, &key, NULL);
	buf[i] -= '\0';
	switch(key) {
	case XK_KP_Enter:
	case XK_Return:
	  change_directory(p->str);
	  return 1;
	}
      }
      break;
    }
    break;
  case EnterNotify:
    switch(p->n) {
    case FILE_PATH:
      break;
    case FILE_NAME:
      strcpy(p->str, current_name);
      break;
    }
    break;
  case LeaveNotify:
    switch(p->n) {
    case FILE_PATH:
      break;
    case FILE_NAME:
      strcpy(current_name, p->str);
      break;
    }
    break;
  }
  return 0;
}

file_background()
{
  if (slotwin[DIRSLOTMAX+1].mode & MenuModeMaskOn)
    scroll_up();
  if (slotwin[DIRSLOTMAX].mode & MenuModeMaskOn)
    scroll_down();
}

/* End of file */
