/* $Header: /ext/yav/catty/xpx/RCS/image.c,v 1.5 1994/04/03 15:46:39 yav Exp $
 */

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <stdio.h>

#include "xpx.h"
#include "extern.h"
#include "scrldef.h"
#include "infodef.h"
#include "cursdef.h"
#include "editdef.h"

#ifndef IMGWINWIDTHMAX
#define IMGWINWIDTHMAX 1024
#endif
#ifndef IMGWINHEIGHTMAX
#define IMGWINHEIGHTMAX 1024
#endif

extern MENU scrlbar[];
extern MENU infomenu[];
extern MENU cursmenu[];
extern GC areagc;

AREA area0;
Bool area_flag = False;

static AREA lastarea;
static AREA current_area;

get_img_window_size()
{
  XWindowAttributes attr;

  XGetWindowAttributes(dsp, win, &attr);
  imgwinwidth = attr.width - IMGX - IMGSCRLW - 16;
  if (imgwinwidth > IMGWINWIDTHMAX)
    imgwinwidth = IMGWINWIDTHMAX;
  imgwinheight = attr.height - IMGY - IMGSCRLW - 16;
  if (imgwinheight > IMGWINHEIGHTMAX)
    imgwinheight = IMGWINHEIGHTMAX;
}

create_img_window()
{
  int i;
  XSetWindowAttributes attr;

  get_img_window_size();
  imgwin = XCreateSimpleWindow(dsp, win, IMGX, IMGY,
			       imgwinwidth, imgwinheight, BDW, fg, bg);
  attr.bit_gravity = NorthWestGravity;
  XChangeWindowAttributes(dsp, imgwin, CWBitGravity, &attr);
  XSelectInput(dsp, imgwin, ExposureMask|
	       ButtonPressMask|ButtonReleaseMask|ButtonMotionMask);
  XMapRaised(dsp, imgwin);
}

image_alloc()
{
  imgdata = (unsigned char *)malloc(imgmaxw*imgmaxh);
  if (imgdata == NULL)
    return 1;
  img = XCreateImage(dsp, DefaultVisual(dsp, scr), DefaultDepth(dsp, scr),
		     ZPixmap, 0, None,
		     imgmaxw*maxzoom, imgmaxh*maxzoom,
		     32, 0);
  if (img == NULL)
    return 1;
  img->data = (char *)malloc(img->bytes_per_line * img->height);
  if (img->data == NULL)
    return 1;
  return 0;
}

init_image()
{
  if (!alloced_color)
    get_color(color_buf+MAXCOLSET*current_pal, MAXCOLSET);
  memset(img->data, 0, img->bytes_per_line * img->height);
}

resize_img_window()
{
  get_img_window_size();
  XResizeWindow(dsp, imgwin, imgwinwidth, imgwinheight);
  resize_scrl();
}

int set_image_size(w, h)
{
  int r;

  r = 0;
  if (w >= 0) {
    imgfilew = w;
    r = 1;
  }
  if (h >= 0) {
    imgfileh = h;
    r |= 2;
  }
  info_imgsize_set();
  return r;
}

void draw_grid(x0, y0, w0, h0)
{
  int n, x, y, w, h;

  n = grid_size*zoomfactor;
  w = w0;
  x = x0;
  if (x & 1) {
    x++;
    w--;
  }
  if (w > 0 && h0 > 0) {
    for (y = y0; y < y0 + h0; y++) {
      if (((imgofsy + y)%n) == 0)
	XDrawLine(dsp, imgwin, gridgc, x, y, x+w-1, y);
    }
  }
  h = h0;
  y = y0;
  if (y & 1) {
    y++;
    h--;
  }
  if (w0 > 0 && h > 0) {
    for (x = x0; x < x0 + w0; x++) {
      if (((imgofsx + x)%n) == 0)
	XDrawLine(dsp, imgwin, gridgc, x, y, x, y+h-1);
    }
  }
}

Bool inarea(x, w, n1, n2)
{
  if (x < n1)
    return (x+w >= n1);
  return x < n2;
}

area_expose(x0, y0, w0, h0)
{
  AREA ar;

  ar = current_area;
  calc_area(&ar.x, &ar.y, &ar.w, &ar.h);
  if ((inarea(ar.x, ar.w, x0, x0+w0) ||
       inarea(ar.y, ar.h, y0, y0+h0)))
    XDrawRectangle(dsp, imgwin, areagc, ar.x, ar.y, ar.w, ar.h);
}

void img_expose(x, y, w, h)
{
  int i;
  
  if ((imgofsx+x >= imgfilew*zoomfactor)||
      (imgofsy+y >= imgfileh*zoomfactor)||
      (imgofsx+x+w < 0)||
      (imgofsy+y+h < 0)) {
    XClearArea(dsp, imgwin, x, y, w, h, False);
    return;
  }
  i = -imgofsx - x;
  if (i > 0) {
    XClearArea(dsp, imgwin, x, y, i, h, False);
    x += i;
    w -= i;
  }
  i = -imgofsy - y;
  if (i > 0) {
    XClearArea(dsp, imgwin, x, y, w, i, False);
    y += i;
    h -= i;
  }
  i = x + w - (imgfilew*zoomfactor - imgofsx);
  if (i > 0) {
    XClearArea(dsp, imgwin, x+w-i, y, i, h, False);
    w -= i;
  }
  i = y + h - (imgfileh*zoomfactor - imgofsy);
  if (i > 0) {
    XClearArea(dsp, imgwin, x, y+h-i, w, i, False);
    h -= i;
  }
  if (w && h) {
    XPutImage(dsp, imgwin, gc0, img, x+imgofsx, y+imgofsy, x, y, w, h);
    if (grid_mode)
      draw_grid(x, y, w, h);
    if (area_flag)
      area_expose(x, y, w, h);
  }
}

image_update_all_sub()
{
  img_expose(0, 0, imgwinwidth, imgwinheight);
  if (view_sw & 1)
    redraw_window(viewwin);
}

void image_update_all()
{
  XFlush(dsp);
  make_view_image();
  gen_image_sub(0, 0, imgfilew, imgfileh);
  image_update_all_sub();
}

void image_update_col()
{
  make_view_image();
  gen_image_col(0, 0, imgfilew, imgfileh);
  image_update_all_sub();
}

void color_update_all()
{
  int i;
  
  free_color();
  get_color(color_buf+MAXCOLSET*current_pal, MAXCOLSET);
  if (alloced_common_color)
    image_update_col();
}

img_scroll(dx, dy)
{
  if (infomenu[INFOMENU_SCROLLSTOP].mode & MenuModeMaskOn) {
    if (dx) {
      if (imgofsx + dx + imgwinwidth >= imgfilew*zoomfactor)
	dx = imgfilew*zoomfactor - (imgofsx + imgwinwidth);
      if (imgofsx + dx < 0)
	dx = -imgofsx;
    }
    if (dy) {
      if (imgofsy + dy + imgwinheight >= imgfileh*zoomfactor)
	dy = imgfileh*zoomfactor - (imgofsy + imgwinheight);
      if (imgofsy + dy < 0)
	dy = -imgofsy;
    }
  }
  if (!dx && !dy)
    return;
  if (dx) {
    imgofsx += dx;
    redraw_window(cursmenu[CURS_TOPX].win);
  }
  if (dy) {
    imgofsy += dy;
    redraw_window(cursmenu[CURS_TOPY].win);
  }
  if (imgwin_noexpose) {
    XCopyArea(dsp, imgwin, imgwin, gc0,
	      (dx<0?0:dx), (dy<0?0:dy),
	      imgwinwidth-(dx<0?-dx:dx), imgwinheight-(dy<0?-dy:dy),
	      (dx<0?-dx:0), (dy<0?-dy:0));
    imgwin_noexpose = False;
    if (dx)
      img_expose((dx<0?0:imgwinwidth-dx), 0, (dx<0?-dx:dx), imgwinheight);
    if (dy)
      img_expose(0, (dy<0?0:imgwinheight-dy), imgwinwidth, (dy<0?-dy:dy));
  } else {
    img_expose(0, 0, imgwinwidth, imgwinheight);
  }
}

imgscrl_background()
{
  int dx, dy;

  /* set default background scroll */
  dx = scroll_dx;
  dy = scroll_dy;
  /* scroll bar press? */
  if (scrlbar[SCRLBAR_TOP].mode & MenuModeMaskOn)
    dy = -scroll_speed;
  if (scrlbar[SCRLBAR_BOTTOM].mode & MenuModeMaskOn)
    dy = scroll_speed;
  if (scrlbar[SCRLBAR_LEFT].mode & MenuModeMaskOn)
    dx = -scroll_speed;
  if (scrlbar[SCRLBAR_RIGHT].mode & MenuModeMaskOn)
    dx = scroll_speed;
  if (dx || dy)
    img_scroll(dx, dy);
}


little_big(p1, p2)
     int *p1;
     int *p2;
{
  int i;
  
  if (*p1 > *p2) {
    i = *p1;
    *p1 = *p2;
    *p2 = i;
  }
}

norm_area(x0, y0, x1, y1)
     int *x0;
     int *y0;
     int *x1;
     int *y1;
{
  little_big(x0, x1);
  little_big(y0, y1);
  *x1 -= *x0;
  *y1 -= *y0;
}

calc_img_area(x, y, w, h)
     int *x;
     int *y;
     int *w;
     int *h;
{
  *w *= zoomfactor;
  *h *= zoomfactor;
  *x = *x * zoomfactor - imgofsx;
  *y = *y * zoomfactor - imgofsy;
}

calc_area(x0, y0, x1, y1)
     int *x0;
     int *y0;
     int *x1;
     int *y1;
{
  little_big(x0, x1);
  little_big(y0, y1);
  *x1 -= *x0;
  *y1 -= *y0;
  calc_img_area(x0, y0, x1, y1);
}

area_exsub(x, y, w, h)
{
  img_expose(x, y, w+1, 1);
  img_expose(x, y, 1, h+1);
  img_expose(x+w, y, 1, h+1);
  img_expose(x, y+h, w+1, 1);
}

copy_area_expose()
{
  AREA ar;
  Bool af;
  
  /* erase old area */
  af = area_flag;
  area_flag = False;
  calc_area(&lastarea.x, &lastarea.y, &lastarea.w, &lastarea.h);
  area_exsub(lastarea.x, lastarea.y, lastarea.w, lastarea.h);
  area_flag = af;
  if (area_flag) {
    ar = lastarea = current_area = area0;
    calc_area(&ar.x, &ar.y, &ar.w, &ar.h);
    area_exsub(ar.x, ar.y, ar.w, ar.h);
  }
}

/* End of file */
