/* $Header: /ext/yav/catty/xpx/RCS/main.c,v 1.32 1994/04/03 15:47:24 yav Exp $
 */

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <stdio.h>

#include "xpx.h"
#include "menudef.h"
#include "editdef.h"
#include "version.h"

#define COLSETX (PALSETX+32)
#define COLSETY PALSETY

#define INTERNAL_ICONPIXMAP 1

/* #define DEFAULT_FONT "8x13" */
/* #define DEFAULT_BOLD_FONT "8x13bold" */
#define DEFAULT_FONT "-adobe-helvetica-medium-r-normal--*-120-*"
#define DEFAULT_BOLD_FONT "-adobe-helvetica-bold-r-normal--*-120-*"

extern MENU colmenu[];
extern char *errname;		/* message.c */

char rcsrev[] = XPX_RCS_REVISION;
char *myname = "xpx";
Display *dsp;
int scr;
Window win;
unsigned long fg, bg, border_color;
GC gc0, gc1, gridgc, areagc;
XImage *img;
Window imgwin;
XFontStruct *fs0, *fs1;
GC colgc[MAXCOLSET];
Window viewwin;
int view_sw = 0;
XImage *vimg = NULL;
int imgmaxw = 640;
int imgmaxh = 400;
int maxzoom = 4;
int current_col = -1;
int current_pal = 0;
int sw_quit = 0;
Bool double_clicked = False;
int zoomfactor = 2;
char imgfilename[MAXPATHLEN] = "*Noname*";
char colfilename[MAXPATHLEN] = "*Noname*";
unsigned char *imgdata;
int imgofsx = 0;
int imgofsy = 0;
int imgwinwidth, imgwinheight;
int alloced_color = 0;
int alloced_common_color = 0;
Bool grid_mode = False;
int grid_size = 8;
int imgfilew = 0;
int imgfileh = 0;
int curx = 0;
int cury = 0;
int curz = 0;
int scroll_speed = 8;
int scroll_dx = 0;
int scroll_dy = 0;
int edit_mode[3] = {EDIT_PLOT, EDIT_GETCOLOR, EDIT_GETCOLOR};
int border_width = 1;
Bool imgwin_noexpose = True;

COL color_buf[MAXPALSET*MAXCOLSET];
char coltbl[MAXPALSET*MAXCOLSET];
unsigned long pixtbl[MAXPALSET*MAXCOLSET];

static char *str_display = "";
static char *str_geometry = "";
static char *str_background = "Wheat";
static char *str_foreground = "Black";
static char *str_border = "Black";
static char *str_font = NULL;
static char *str_fontbold = NULL;
static Bool use_private_color = True;
static char grid_dash_list[2] = {1,1};
static Pixmap icon_pixmap = None;
#ifdef INTERNAL_ICONPIXMAP
#include "icon.xbm"
#endif

usage()
{
  fprintf(stderr, "* %s %s *\n", myname, rcsrev);
  fprintf(stderr, "written by yav (HBA00622@niftyserve.or.jp UHD98984@pcvan.or.jp)\n");
  fprintf(stderr, "usage : %s [options] col.kcf rei.cel\n", myname);
  exit(1);
}

char *filename(name)
     char *name;
{
  char *p;
  extern char *strrchr();

  if (name[0] == '\0')
    return "*Noname*";
  p = strrchr(name, '/');
  if (p != NULL)
    name = p+1;
  return name;
}

char *basefilename(name)
     char *name;
{
  char *p;
  extern char *strrchr();

  name = filename(name);
  p = strrchr(name, '.');
  if (p != NULL)
    *p = '\0';
  return name;
}


#define col8to32(x) (((x)<<24)|((x)<<16)|((x)<<8)|(x))
calcol(col, p)
     XColor *col;
     COL *p;
{
  col->red = col8to32(p->rgb.red);
  col->green = col8to32(p->rgb.green);
  col->blue = col8to32(p->rgb.blue);
  col->flags = DoRed|DoGreen|DoBlue;
}

get_color_sub(i, p)
     COL *p;
{
  XColor col;
  unsigned long plane_masks[1];
  int r;

  r = 0;
  calcol(&col, p);
  if (coltbl[i] & CM_PRIVATE) {
    col.pixel = pixtbl[i];
    XStoreColor(dsp, DefaultColormap(dsp, scr), &col);
  } else {
    if (use_private_color &&
	XAllocColorCells(dsp, DefaultColormap(dsp, scr), False,
			 plane_masks, 0, &col.pixel, 1)) {
      XStoreColor(dsp, DefaultColormap(dsp, scr), &col);
      coltbl[i] = CM_ALLOCED | CM_UPDATING | CM_PRIVATE;
      alloced_color++;
    } else if (XAllocColor(dsp, DefaultColormap(dsp, scr), &col)) {
      coltbl[i] = CM_ALLOCED | CM_UPDATING;
      alloced_common_color++;
      alloced_color++;
    } else {
      r = 1;			/* alloc error! */
      col.pixel = bg;
    }
    pixtbl[i] = col.pixel;
  }
  return r;
}

get_color(p, n)
     COL *p;
{
  int i;

  for (i = 0; i < n; i++) {
    if (get_color_sub(i, p))
      message("!color %d %d %d cannot allocated!\n",
	      p->rgb.red, p->rgb.green, p->rgb.blue);
    if (i < MAXCOLSET) {
      XSetForeground(dsp, colgc[i], pixtbl[i]);
      update_colmenu(i);
    }
    p++;
  }
  return i;
}

/* return number of allocated color */
get_requested_colors()
{
  int i, r;
  
  r = 0;
  for (i = 0; i < MAXCOLSET*MAXPALSET; i++) {
    if (coltbl[i] & CM_REQUEST) {
      if (get_color_sub(i, &color_buf[i]) == 0)
	r++;
      coltbl[i] &= ~CM_REQUEST;
    }
  }
  return r;
}

request_color(p)
     COL *p;
{
  int i;
  
  for (i = 0; i < MAXCOLSET*MAXPALSET; i++) {
    if (!(coltbl[i] & (CM_ALLOCED|CM_REQUEST))) {
      coltbl[i] = CM_REQUEST;
      color_buf[i].rgb.red = p->rgb.red;
      color_buf[i].rgb.green = p->rgb.green;
      color_buf[i].rgb.blue = p->rgb.blue;
      return i;
    }
  }
  return -1;
}

request_color2(i, p)
     COL *p;
{
  coltbl[i] = CM_REQUEST;
  color_buf[i].rgb.red = p->rgb.red;
  color_buf[i].rgb.green = p->rgb.green;
  color_buf[i].rgb.blue = p->rgb.blue;
}

init_color()
{
  alloced_color = 0;
  memset(coltbl, 0, sizeof(coltbl));
}

/* free common color cells */
free_color()
{
  int i;

  for (i = 0; i < MAXPALSET*MAXCOLSET; i++) {
    if ((coltbl[i] & CM_ALLOCED)&&!(coltbl[i] & CM_PRIVATE)) {
      XFreeColors(dsp, DefaultColormap(dsp, scr), pixtbl+i, 1, 0);
      --alloced_color;
      coltbl[i] = 0;
    }
  }
}

/* free common and private color cells */
free_color_all()
{
  int i;

  for (i = 0; i < MAXPALSET*MAXCOLSET; i++) {
    if (coltbl[i] & CM_ALLOCED) {
      XFreeColors(dsp, DefaultColormap(dsp, scr), pixtbl+i, 1, 0);
      --alloced_color;
      coltbl[i] = 0;
    }
  }
}

init_color_buf()
{
  int i, j;
  typedef struct {
    unsigned char r;
    unsigned char g;
    unsigned char b;
  } RGBDATA;
  static RGBDATA initdata[16] = {
    {  0,  0,  0},{  0,  0,255},{255,  0,  0},{255,  0,255},
    {  0,255,  0},{  0,255,255},{255,255,  0},{127,127,127},
    { 64, 64, 64},{127,127,255},{255,127,127},{255,127,255},
    {127,255,127},{127,255,255},{255,255,127},{255,255,255}
  }; 
#define GRAY(n) {n*17, n*17, n*17}
  static RGBDATA initdata1[16] = {
    GRAY(15), GRAY(0), GRAY(1), GRAY(2),
    GRAY(3), GRAY(4), GRAY(5), GRAY(6),
    GRAY(7), GRAY(8), GRAY(9), GRAY(10),
    GRAY(11), GRAY(12), GRAY(13), GRAY(14)
    };
  static RGBDATA initdata2[16] = {
    GRAY(0), GRAY(1), GRAY(2), GRAY(3),
    GRAY(4), GRAY(5), GRAY(6), GRAY(7),
    GRAY(8), GRAY(9), GRAY(10), GRAY(11),
    GRAY(12), GRAY(13), GRAY(14), GRAY(15)
    };
  
  for (i = 0; i < MAXPALSET; i++) {
    for (j = 0; j < MAXCOLSET; j++) {
      color_buf[MAXCOLSET*i+j].rgb.padding = 0;
      color_buf[MAXCOLSET*i+j].rgb.blue = initdata[j].b;
      color_buf[MAXCOLSET*i+j].rgb.red = initdata[j].r;
      color_buf[MAXCOLSET*i+j].rgb.green = initdata[j].g;
    }
  }
  for (j = 0; j < MAXCOLSET; j++) {
    color_buf[MAXCOLSET+j].rgb.padding = 0;
    color_buf[MAXCOLSET+j].rgb.blue = initdata1[j].b;
    color_buf[MAXCOLSET+j].rgb.red = initdata1[j].r;
    color_buf[MAXCOLSET+j].rgb.green = initdata1[j].g;
  }
  for (j = 0; j < MAXCOLSET; j++) {
    color_buf[MAXCOLSET*2+j].rgb.padding = 0;
    color_buf[MAXCOLSET*2+j].rgb.blue = initdata2[j].b;
    color_buf[MAXCOLSET*2+j].rgb.red = initdata2[j].r;
    color_buf[MAXCOLSET*2+j].rgb.green = initdata2[j].g;
  }
#undef GRAY
}

key_event(ev)
     XEvent *ev;
{
  int i;
  char buf[32];
  KeySym key;
  XComposeStatus cs;
  
  i = XLookupString((XKeyEvent *)ev, buf, sizeof(buf)-1, &key, &cs);
  buf[i] = '\0';
  switch(buf[0]) {
  case 0x03:
  case 'q':
    menu_stat_change(MENU_QUIT, 2);
    break;
  case 'f':
    menu_stat_change(MENU_FILE, 2);
    break;
  case 'i':
    menu_stat_change(MENU_INFO, 2);
    break;
  case 'e':
    menu_stat_change(MENU_EDIT, 2);
    break;
  case '-':
    zoom_out();
    break;
  case '+':
    zoom_in();
    break;
  case 'L'-0x40:
    redraw_window(imgwin);
    break;
  case '2':
  case 'j':
    img_scroll(0, scroll_speed);
    break;
  case '8':
  case 'k':
    img_scroll(0, -scroll_speed);
    break;
  case '6':
  case 'l':
    img_scroll(scroll_speed, 0);
    break;
  case '4':
  case 'h':
    img_scroll(-scroll_speed, 0);
    break;
  case '1':
    img_scroll(-scroll_speed, scroll_speed);
    break;
  case '3':
    img_scroll(scroll_speed, scroll_speed);
    break;
  case '7':
    img_scroll(-scroll_speed, -scroll_speed);
    break;
  case '9':
    img_scroll(scroll_speed, -scroll_speed);
    break;
  case 'p':
    edit_mode_change(EDIT_PLOT);
    break;
  case 'c':
    edit_mode_change(EDIT_COPY);
    break;
  case 'b':
    edit_mode_change(EDIT_FILL);
    break;
  }
}

string_option(ac, av, r)
     int *ac; char ***av; char **r;
{
  if (*ac > 1) {
    --*ac;
    ++*av;
    *r = **av;
  }
}

f_display(ac, av)
     int *ac; char ***av;
{
  string_option(ac, av, &str_display);
}

f_geometry(ac, av)
     int *ac; char ***av;
{
  string_option(ac, av, &str_geometry);
}

f_background(ac, av)
     int *ac; char ***av;
{
  string_option(ac, av, &str_background);
}

f_foreground(ac, av)
     int *ac; char ***av;
{
  string_option(ac, av, &str_foreground);
}

f_bordercolor(ac, av)
     int *ac; char ***av;
{
  string_option(ac, av, &str_border);
}

f_font(ac, av)
     int *ac; char ***av;
{
  string_option(ac, av, &str_font);
}

f_fontbold(ac, av)
     int *ac; char ***av;
{
  string_option(ac, av, &str_fontbold);
}

f_commoncolor(ac, av)
     int *ac; char ***av;
{
  use_private_color = False;
}

numeric_option(ac, av, r)
     int *ac; char ***av; int *r;
{
  int i;
  char *p;
  
  if (*ac < 1)
    return 0;
  --*ac;
  ++*av;
  i = strtol(**av, &p, 0);
  if (p != **av)
    *r = i;
  return p - **av;
}

f_borderwidth(ac, av)
     int *ac; char ***av;
{
  numeric_option(ac, av, &border_width);
}

f_maxw(ac, av)
     int *ac; char ***av;
{
  numeric_option(ac, av, &imgmaxw);
}

f_maxh(ac, av)
     int *ac; char ***av;
{
  numeric_option(ac, av, &imgmaxh);
}

f_maxz(ac, av)
     int *ac; char ***av;
{
  numeric_option(ac, av, &maxzoom);
  if (zoomfactor > maxzoom)
    zoomfactor = maxzoom;
}

f_scrollspeed(ac, av)
     int *ac; char ***av;
{
  numeric_option(ac, av, &scroll_speed);
}

struct {
  char *name;
  int (*func)();
} opttbl[] = {
  {"-display", f_display},
  {"-geometry", f_geometry},
  {"-background", f_background},
  {"-foreground", f_foreground},
  {"-bordercolor", f_bordercolor},
  {"-borderwidth", f_borderwidth},
  {"-font", f_font},
  {"-bg", f_background},
  {"-fg", f_foreground},
  {"-bd", f_bordercolor},
  {"-bw", f_borderwidth},
  {"-fn", f_font},
  {"-fb", f_fontbold},
  {"-commoncolor", f_commoncolor},
  {"-maxw", f_maxw},
  {"-maxh", f_maxh},
  {"-maxz", f_maxz},
  {"-scrollspeed", f_scrollspeed},
  {"-?", usage},
  {"-h", usage},
  {"-V", usage},
  {"-help", usage},
  {NULL, NULL}
};

create_icon()
{
#ifdef INTERNAL_ICONPIXMAP
  icon_pixmap =
#if 1
    XCreateBitmapFromData(dsp, DefaultRootWindow(dsp), 
			  icon_bits, icon_width, icon_height);
#else
  XCreatePixmapFromBitmapData(dsp, DefaultRootWindow(dsp), 
			      icon_bits, icon_width, icon_height,
			      BlackPixel(dsp, scr), WhitePixel(dsp, scr),
			      DefaultDepth(dsp, scr));
#endif
  if (icon_pixmap == None) {
    error("icon pixmap create error!\n");
    return 1;
  }
#endif /* INTERNAL_ICONPIXMAP */
  return create_icon_pixmap();
}

free_icon()
{
  free_icon_pixmap();
  if (icon_pixmap != None)
    XFreePixmap(dsp, icon_pixmap);
}

parse_color(col, name)
     XColor *col;
     char *name;
{
  if (!XParseColor(dsp, DefaultColormap(dsp, scr), name, col)||
      !XAllocColor(dsp, DefaultColormap(dsp, scr), col))
    return error("``%s'' color alloc error!\n", name);
  return 0;
}

main(argc, argv)
     int argc; char **argv;
{
  XSizeHints hint;
  XEvent ev;
  int oargc;
  char **oargv;
  int i;
  Window double_click_win = 0;
  Time l, double_click_time;
  XColor col;
  XEvent ahead;
  Bool ignore_event;

  oargc = argc;
  oargv = argv;
  errname = myname = filename(*argv);
  while (--argc && **++argv == '-') {
    for (i = 0; opttbl[i].name != NULL; i++) {
      if (strcmp(*argv, opttbl[i].name) == 0) {
	opttbl[i].func(&argc, &argv);
	break;
      }
    }
    if (opttbl[i].name == NULL && *argv)
      return error("unknown option ``%s''.\n", *argv);
  }
  dsp = XOpenDisplay(str_display);
  if (!dsp) {
    error("Cannot open display!\n");
    exit(1);
  }
  scr = DefaultScreen(dsp);
  bg = WhitePixel(dsp, scr);
  border_color = fg = BlackPixel(dsp, scr);
  if (parse_color(&col, str_background) == 0)
    bg = col.pixel;
  if (parse_color(&col, str_foreground) == 0)
    fg = col.pixel;
  if (parse_color(&col, str_border) == 0)
    border_color = col.pixel;
  create_icon();
  hint.flags = 0;
  i = XGeometry(dsp, scr, str_geometry, "428x368", border_width, 1, 1, 0, 0,
		&(hint.x), &(hint.y), &(hint.width), &(hint.height));
  if (i & (XValue|YValue))
    hint.flags |= PPosition;
  if (i & (WidthValue|HeightValue))
    hint.flags |= PSize;
  win = XCreateSimpleWindow(dsp, DefaultRootWindow(dsp),
			    hint.x, hint.y, hint.width, hint.height,
			    border_width, border_color, bg);
  XSetStandardProperties(dsp, win, *oargv, *oargv, icon_pixmap,
			 oargv, oargc, &hint);

  if (str_fontbold == NULL)
    str_fontbold = str_font == NULL ? DEFAULT_BOLD_FONT : str_font;
  if (str_font == NULL)
    str_font = DEFAULT_FONT;

  fs0 = XLoadQueryFont(dsp, str_font);
  fs1 = XLoadQueryFont(dsp, str_font);

  gc0 = XCreateGC(dsp, win, 0, 0);
  XSetBackground(dsp, gc0, bg);
  XSetForeground(dsp, gc0, fg);
  XSetFont(dsp, gc0, fs0->fid);

  gc1 = XCreateGC(dsp, win, 0, 0);
  XSetBackground(dsp, gc1, fg);
  XSetForeground(dsp, gc1, bg);
  XSetFont(dsp, gc1, fs1->fid);

  areagc = XCreateGC(dsp, win, 0, 0);
  XSetBackground(dsp, areagc, bg);
  XSetForeground(dsp, areagc, fg);

  gridgc = XCreateGC(dsp, win, 0, 0);
  XSetBackground(dsp, gridgc, bg);
  XSetForeground(dsp, gridgc, bg);
  XSetLineAttributes(dsp, gridgc, 0, LineOnOffDash, CapButt, JoinMiter);
  XSetDashes(dsp, gridgc, 0, grid_dash_list, 2);

  create_img_window();

  if (!init_file_type())
    exit(1);
  menuev_init();
  create_message_window();
  create_info_window();
  create_scrlbar_window();
  create_menu_window(MENUX, MENUY);
  create_colset_window(COLSETX, COLSETY);
  create_palset_window(PALSETX, PALSETY);
  create_filename_window();
  create_zoom_window(256, 4);
  create_grid_window(340, 4);
  create_cursor_window(256, 32);
  create_edit_window();
  create_file_window();
  if (image_alloc()||undo_init()) {
    error("more core!");
    exit(1);
  }
  
  init_color_buf();
  init_color();
  init_image();
  image_update_all();
  while (argc--)
    read_data_file(*argv++);
  XSelectInput(dsp, win, ExposureMask|StructureNotifyMask|
	       ButtonPressMask|KeyPressMask);
  XMapRaised(dsp, win);
  while (!sw_quit) {
    if (XEventsQueued(dsp, QueuedAfterReading) == 0) {
      XFlush(dsp);
      imgscrl_background();
      file_background();
    }
    XNextEvent(dsp, &ev);
    ignore_event = False;
    switch(ev.type) {
    case NoExpose:
      if (ev.xnoexpose.drawable == imgwin)
	imgwin_noexpose = True;
      break;
    case GraphicsExpose:
      if (ev.xgraphicsexpose.drawable == imgwin) {
	img_expose(ev.xgraphicsexpose.x, ev.xgraphicsexpose.y,
		   ev.xgraphicsexpose.width, ev.xgraphicsexpose.height);
	if (ev.xgraphicsexpose.count == 0)
	  imgwin_noexpose = True;
      }
      break;
    case Expose:
      if (ev.xexpose.window == imgwin)
	img_expose(ev.xexpose.x, ev.xexpose.y,
		   ev.xexpose.width, ev.xexpose.height);
      else
	info_expose(&ev);
      break;
    case MappingNotify:
      XRefreshKeyboardMapping((XMappingEvent *)&ev);
      break;
    case KeyPress:
      if (ev.xkey.window == win)
	key_event(&ev);
      break;
    case ConfigureNotify:
      if (ev.xconfigure.window == win)
	resize_img_window();
      break;
    case ButtonPress:
      if (ev.xbutton.window == double_click_win) {
	l = ev.xbutton.time - double_click_time;
	if (l & 0x80000000)
	  l = ~l + 1;
	double_clicked = l < 500;
      } else {
	double_clicked = False;
      }
      double_click_win = ev.xbutton.window;
      double_click_time = ev.xbutton.time;
      info_press(&ev);
      break;
    case EnterNotify:
      if (XEventsQueued(ev.xcrossing.display, QueuedAfterReading) > 0) {
	XPeekEvent(ev.xcrossing.display, &ahead);
	if (ahead.type == LeaveNotify &&
	    ahead.xcrossing.window == ev.xcrossing.window &&
	    ahead.xcrossing.mode == ev.xcrossing.mode &&
	    ahead.xcrossing.detail == ev.xcrossing.detail) {
	  XNextEvent(ev.xcrossing.display, &ev);
	  ignore_event = True;
	}
      }
    case LeaveNotify:
      double_click_win = 0;
      break;
    }
    if (!ignore_event) {
      edit_event(&ev);
      menuev_event(&ev);
    }
  }
  menuev_destroy();
  close_message_window();
  close_edit_window();
  close_info_window();
  close_file_window();
  XDestroyWindow(dsp, imgwin);
  free_color_all();
  free_icon();
  XFree((char *)img);
  XFreeGC(dsp, gc1);
  XFreeGC(dsp, gc0);
  XDestroyWindow(dsp, win);
  XCloseDisplay(dsp);
  exit(0);
}

/* End of file */
