/* $Header: /ext/yav/catty/xpx/RCS/message.c,v 1.3 1994/03/17 17:39:45 yav Exp $
 */

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <stdio.h>
#ifdef SYSV
#include <varargs.h>
#else
#include <stdarg.h>
#endif /* SYSV */

#include "xpx.h"
#include "extern.h"

#define MSGLINEMAX	16
#define MSGLINESPACE	20
#define MSGWIDTH	400
#define MSGHEIGHT	(MSGLINESPACE*MSGLINEMAX)

char *errname = NULL;
int errcode = 1;
static Window messagewin;
static MENU msgline[MSGLINEMAX+1];

int error(
#ifdef __STDC__
	const char *format, ...)
#else
     format)
     char *format;
#endif
{
#ifdef __STDC__
  va_list ap;
  va_start(ap, format);
#endif
  if (errname != NULL)
    fprintf(stderr, " %s:", errname);
#ifdef __STDC__
  vfprintf(stderr, format, ap);
#else
  vfprintf(stderr, format, &format + 1);
#endif
  fprintf(stderr, "\007\n");
#ifdef __STDC__
  va_end(ap);
#endif
  return errcode;
}

static int current_line = 0;

message_scroll()
{
  int i;
  char *s;

  s = msgline[0].str;
  for (i = 0; i < MSGLINEMAX-1; i++) {
    msgline[i].str = msgline[i+1].str;
    redraw_window(msgline[i].win);
  }
  *s = '\0';
  msgline[i].str = s;
}

void message_store(str)
     char *str;
{
  char *p;
  extern char *strchr();

  for (;;) {
    p = strchr(str, '\n');
    if (p != NULL)
      *p++ = '\0';
    strncat(msgline[current_line].str, str, msgline[current_line].strlen);
    redraw_window(msgline[current_line].win);
    if (p == NULL)
      break;
    /* new line */
    if (++current_line >= MSGLINEMAX) {
      message_scroll();
      current_line = MSGLINEMAX-1;
    }
    str = p;
  }
}

void message(
#ifdef __STDC__
	const char *format, ...)
#else
     format)
     char *format;
#endif
{
#ifdef __STDC__
  va_list ap;
#endif
  char buf[1024];
  
#ifdef __STDC__
  va_start(ap, format);
#endif
  if (*format == '!') {
    map_message_window();
    format++;
  }
#ifdef __STDC__
  vsprintf(buf, format, ap);
#else
  vsprintf(buf, format, &format + 1);
#endif
  message_store(buf);
#ifdef __STDC__
  va_end(ap);
#endif
}

int msgev();

create_message_window()
{
  int i, x, y;
  MENU *p;
  XSizeHints hint;
  char buf[256];
  
  get_drawable_size(DefaultRootWindow(dsp), &x, &y);
  hint.x = (x-MSGWIDTH)/2;
  hint.y = (y-MSGHEIGHT)/2;
  hint.width = MSGWIDTH; hint.height = MSGHEIGHT;
  hint.flags = USPosition|USSize;
  messagewin = XCreateSimpleWindow(dsp, DefaultRootWindow(dsp), hint.x, hint.y,
				   hint.width, hint.height,
				   border_width, border_color, bg);
  sprintf(buf, "%s message", myname);
  XSetStandardProperties(dsp, messagewin, buf, buf, None, NULL, NULL, &hint);
  XSelectInput(dsp, messagewin, ExposureMask);

  p = msgline;
  y = 0;
  for (i = 0; i < MSGLINEMAX; i++) {
    memset(p, 0, sizeof(*p));
    p->n = i;
    p->y = y;
    p->w = MSGWIDTH+400;
    p->h = MSGLINESPACE;
    p->mode = MenuModeMaskLeft;
    p->im = ExposureMask|ButtonPressMask;
    p->event = msgev;
    p->strlen = 64;
    p++;
    y += MSGLINESPACE;
  }
  p->n = -1;
  menuwin_create(msgline, messagewin, 0, 0);
  return 0;
}

map_message_window()
{
  XMapRaised(dsp, messagewin);
}

unmap_message_window()
{
  XUnmapWindow(dsp, messagewin);
}

close_message_window()
{
  XDestroyWindow(dsp, messagewin);
}

int msgev(p, ev)
     MENU *p;
     XEvent *ev;
{
  switch(ev->type) {
  case ButtonPress:
    unmap_message_window();
    return 1;
  }
  return 0;
}

/* End of file */
