/* $Header: /ext/yav/catty/xpx/RCS/sub.c,v 1.15 1994/04/03 15:48:43 yav Exp $
 */

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <stdio.h>

#include "xpx.h"
#include "extern.h"
#include "icondef.h"

extern char *malloc();

#define MENUTBLMAX 64
static MENU *menutbl[MENUTBLMAX];

get_drawable_size(dr, wp, hp)
     Drawable dr;
     int *wp;
     int *hp;
{
  Window root;
  int x, y;
  unsigned int width, height, border, depth;
  
  XGetGeometry(dsp, dr, &root, &x, &y, &width, &height, &border, &depth);
  if (wp != NULL)
    *wp = width;
  if (hp != NULL)
    *hp = height;
}

MENU *get_top_menu(p)
     MENU *p;
{
  while (p->n)
    --p;
  return p;
}

#define highlightGC(mode) \
  (((mode) & MenuModeMaskEnter)?(((mode) & MenuModeMaskOn)?gc1:gc0):\
   (((mode) & MenuModeMaskOn)?gc0:gc1))

menu_highlight(p)
     MENU *p;
{
  XDrawRectangle(dsp, p->win, highlightGC(p->mode), 0, 0, p->w-1, p->h-1);
}

menuwin_strpos_center(wd, fs, str, xp, yp, lp, ascp, desp, w, h)
     Window wd;
     XFontStruct *fs;
     char *str;
     int *xp;
     int *yp;
     int *lp;
     int *ascp;
     int *desp;
     int w;
     int h;
{
  int len, tw, dir, asc, des;
  XCharStruct overall;
  
  len = strlen(str);
  tw = XTextWidth(fs, str, len);
  XTextExtents(fs, str, 0, &dir, &asc, &des, &overall);
  if (xp != NULL)
    *xp = (w-tw)/2;
  if (yp != NULL)
    *yp = h - (h-(asc+des))/2 - des;
  if (lp != NULL)
    *lp = len;
  if (ascp != NULL)
    *ascp = asc;
  if (desp != NULL)
    *desp = des;
}

#define menu_fill_black(menup) \
  XFillRectangle(dsp,(menup)->win,gc0,0,0,(menup)->w,(menup)->h)

draw_center_string(wd, w, h, fs, gc, str)
     Window wd;
     int w;
     int h;
     XFontStruct *fs;
     GC gc;
     char *str;
{
  int x, y, len;
  
  menuwin_strpos_center(wd, fs, str, &x, &y, &len, NULL, NULL, w, h);
  XDrawImageString(dsp, wd, gc, x, y, str, len);
}

menuwin_center(wd, w, h, str)
     Window wd;
     int w;
     int h;
     char *str;
{
  draw_center_string(wd, w, h, fs0, gc0, str);
}

menu_center(p, str)
     MENU *p;
     char *str;
{
  XFontStruct *fs;
  GC gc;

  if (p->mode & MenuModeMaskOn) {
    menu_fill_black(p);
    gc = gc1;
  } else {
    gc = gc0;
  }
  draw_center_string(p->win, p->w, p->h, fs0, gc, str);
  if (p->im & (KeyPressMask|ButtonPressMask))
    menu_highlight(p);
}


menu_left(p, str)
     MENU *p;
     char *str;
{
  GC gc;
  
  if (p->mode & MenuModeMaskOn) {
    menu_fill_black(p);
    gc = gc1;
  } else {
    gc = gc0;
  }
  XDrawImageString(dsp, p->win, gc, 2, p->h-5, str, strlen(str));
  if (p->im & (ButtonPressMask|KeyPressMask))
    menu_highlight(p);
}

draw_cursor(p, str)
     MENU *p;
     char *str;
{
  int x, y, len, asc, des;

  menuwin_strpos_center(p->win, fs0, str, &x, &y,
			&len, &asc, &des, p->w, p->h);
  if (len)
    x += XTextWidth(fs0, str, p->cursor);
  XDrawLine(dsp, p->win, gc0, x, y-asc, x, y);
  XDrawLine(dsp, p->win, gc0, x, y, x-2, y+1);
  XDrawLine(dsp, p->win, gc0, x, y, x+2, y+1);
  XDrawLine(dsp, p->win, gc0, x, y-asc, x-2, y-asc-1);
  XDrawLine(dsp, p->win, gc0, x, y-asc, x+2, y-asc-1);
}

menu_icon_center(p, icon)
     MENU *p;
{
  if (p->mode & MenuModeMaskOn) {
    menu_fill_black(p);
    icon |= ICON_REVERSE;
  }    
  XCopyArea(dsp, icon_pixmap(icon), p->win, gc0,
	    0, 0, icon_width(icon), icon_height(icon),
	    (p->w-icon_width(icon))/2, (p->h-icon_height(icon))/2);
  menu_highlight(p);
}

menuev_init()
{
  int i;
  
  for (i = 0; i < MENUTBLMAX; i++)
    menutbl[i] = NULL;
}

set_menutbl(menu)
     MENU *menu;
{
  int i;

  for (i = 0; i < MENUTBLMAX; i++) {
    if (menutbl[i] == NULL) {
      menutbl[i] = menu;
      return 0;
    }
  }
  return 1;
}

unset_menutbl(menu)
     MENU *menu;
{
  int i;
  
  for (i = 0; i < MENUTBLMAX; i++) {
    if (menutbl[i] == menu) {
      menutbl[i] = NULL;
      return 0;
    }
  }
  return 1;
}

menuwin_create(menu, parent, x0, y0)
     MENU *menu;
     Window parent;
{
  MENU *p;
  XSetWindowAttributes attr;

  for (p = menu; p->n >= 0; p++) {
    if (p->strlen && p->str == NULL) {
      p->str = malloc(p->strlen);
      if (p->str == NULL)
	error("menuwin_create: more core!");
      else
	*(p->str) = '\0';
    }
    p->win = 
      XCreateSimpleWindow(dsp, parent, x0 + p->x, y0 + p->y, p->w, p->h,
			  p->mode & MenuModeMaskBorder ? BDW : 0, fg, bg);
    XSelectInput(dsp, p->win, p->im);
    if (p->gr) {
      attr.win_gravity = p->gr;
      XChangeWindowAttributes(dsp, p->win, CWWinGravity, &attr);
    }
    XMapRaised(dsp, p->win);
  }
  set_menutbl(menu);
}

menuwin_destroy(menu)
     MENU *menu;
{
  MENU *p;

  for (p = menu; p->n >= 0; p++)
    XDestroyWindow(dsp, p->win);
  unset_menutbl(menu);
}

menuwin_expose_default(p)
     MENU *p;
{
  char *s;
  
  s = p->strfunc ? p->strfunc(p) : p->str;
  if (s == NULL)
    return;
  if (p->mode & MenuModeMaskLeft)
    menu_left(p, s);
  else
    menu_center(p, s);
  /* if highlight on , display input cursor */
  if ((p->im & KeyPressMask)&&(p->mode & MenuModeMaskEnter))
    draw_cursor(p, s);
}

redraw_window(w)
     Window w;
{
  XClearArea(dsp, w, 0, 0, 0, 0, True);
}

menuev_destroy()
{
  int i;
  
  for (i = 0; i < MENUTBLMAX; i++) {
    if (menutbl[i] != NULL)
      menuwin_destroy(menutbl[i]);
  }
}

menuev_event_sub(p, ev)
     MENU *p;
     XEvent *ev;
{
  void (*func)();
  short last_mode;
  
  if (p->event != NULL && p->event(p, ev))
    return;
  /* default event operation */
  last_mode = p->mode;
  switch(ev->type) {
  case Expose:
    if (!ev->xexpose.count)
      menuwin_expose_default(p);
    return;
  case GraphicsExpose:
    if (!ev->xgraphicsexpose.count)
      menuwin_expose_default(p);
    return;
  case NoExpose:
    return;
  case ButtonPress:
    if (p->im & ButtonPressMask) {
      if (p->im & ButtonReleaseMask)
	p->mode |= MenuModeMaskOn; /* select on */
      else
	p->mode ^= MenuModeMaskOn; /* select alternate */
    }
    break;
  case ButtonRelease:
    if (p->im & ButtonReleaseMask)
      p->mode &= ~MenuModeMaskOn; /* select off */
    break;
  case EnterNotify:
    if (p->im & (ButtonPressMask|KeyPressMask)) {
      p->mode |= MenuModeMaskEnter; /* highlight on */
      if (!(p->im & KeyPressMask)) {
	menu_highlight(p);
	return;
      }
    }
    break;
  case LeaveNotify:
    if (ev->xcrossing.detail == NotifyInferior)
      return;
    if (p->im & (ButtonPressMask|KeyPressMask)) {
      p->mode &= ~MenuModeMaskEnter; /* highlight off */
      if (!(p->im & KeyPressMask)) {
	menu_highlight(p);
	return;
      }
    }
    break;
  case KeyPress:
    edit_string(ev, p->strfunc?p->strfunc(p):p->str,p->strlen, &p->cursor);
    redraw_window(p->win);
    return;
  }
  if (p->mode & ~MenuModeMaskEnter == last_mode & ~MenuModeMaskEnter)
    menuwin_expose_default(p);
  else
    redraw_window(p->win);
}

Drawable get_event_window(ev)
     XEvent *ev;
{
  Drawable win;

  win = 0;
  switch(ev->type) {
  case Expose:
    win = ev->xexpose.window;
    break;
  case ButtonPress:
    win = ev->xbutton.window;
    break;
  case ButtonRelease:
    win = ev->xbutton.window;
    break;
  case EnterNotify:
    win = ev->xcrossing.window;
    break;
  case LeaveNotify:
    win = ev->xcrossing.window;
    break;
  case KeyPress:
    win = ev->xkey.window;
    break;
  case GraphicsExpose:
    win = ev->xgraphicsexpose.drawable;
    break;
  case NoExpose:
    win = ev->xnoexpose.drawable;
    break;
  case MotionNotify:
    win = ev->xmotion.window;
    break;
  }
  return win;
}

menuev_event(ev)
     XEvent *ev;
{
  int i;
  MENU *p;
  Drawable win;

  win = get_event_window(ev);
  if (!win)
    return;
  for (i = 0; i < MENUTBLMAX; i++) {
    for (p = menutbl[i]; p != NULL && p->n >= 0; p++) {
      if (p->win == win) {
	menuev_event_sub(p, ev);
	return;
      }
    }
  }
}

/* End of file */
